from ._better_print import better_print as BetterPrint


def better_input(*args, sep: str = ' ', delay: float = .01) -> str:
	r"""
	Prints args seperated by sep with the given delay between each character and gets an input
	:param args:
	:param sep:
	:param delay:
	:return: str
	"""

	if not args:
		text = ''
	else:
		text = sep.join(args)

	if text:
		BetterPrint(text, delay=delay, end='')
	res = input()
	return res



from typing import Callable
from .Terminal.color import foreground

RESET = foreground.RESET.value
numberColor = RESET + foreground.YELLOW.value
boolColor = RESET + foreground.LBLUE.value
stringColor = RESET + foreground.LGREEN.value
quotesColor = RESET + foreground.BOLD.value + foreground.CYAN.value

type Iterable = dict | list | tuple | set


def _f(x: object):
	if isinstance(x, str):
		quotes = '"' if "\n" not in x else '"""'
		return f"{quotesColor}{quotes}{stringColor}{x.replace("\"", "\\\"")}{quotesColor}{quotes}{RESET}"
	elif isinstance(x, (bool, type(None))):
		return f"{boolColor}{x}{RESET}"
	elif isinstance(x, (int, float)):
		return f"{numberColor}{x}{RESET}"
	return x


def _dictOut(_dict: dict, indentationLevel: int, indentationString: str) -> str:
	indent = indentationString * indentationLevel
	output = ""
	for k, v in _dict.items():
		if isinstance(v, (dict, list, tuple, set)):
			output += f"{indent}{_f(k)}: "
			output += _iterableOut(v, indentationLevel, indentationString, indentBefore=False)
		else:
			output += f"{indent}{_f(k)}: {_f(v)}\n"
	return output


def _listOut(_list: list, indentationLevel: int, indentationString: str) -> str:
	indent = indentationString * indentationLevel
	output = ""
	for item in _list:
		if isinstance(item, (dict, list, tuple, set)):
			output += _iterableOut(item, indentationLevel, indentationString)
		else:
			output += f"{indent}{_f(item)}\n"
	return output


def _tupleOut(_tuple: tuple, indentationLevel: int, indentationString: str) -> str:
	return _listOut(list(_tuple), indentationLevel, indentationString)


def _setOut(_set: set, indentationLevel: int, indentationString: str) -> str:
	return _listOut(list(_set), indentationLevel, indentationString)


def _callFunc(
		brackets: str,
		func: Callable[..., str],
		iterable: Iterable,
		indentationLevel: int,
		indentationString: str,
		indentBefore: bool) -> str:
	indent = indentationString * indentationLevel
	return (f"{indent if indentBefore else ""}{brackets[0]}\n"
	        f"{func(iterable, indentationLevel + 1, indentationString)}{indent}{brackets[1]}\n")


def _iterableOut(iterable: Iterable, indentationLevel: int, indentationString: str, *,
                 indentBefore: bool = True) -> str:
	output = ""
	if isinstance(iterable, dict):
		output += _callFunc("{}", _dictOut, iterable, indentationLevel, indentationString, indentBefore)
	elif isinstance(iterable, list):
		output += _callFunc("[]", _listOut, iterable, indentationLevel, indentationString, indentBefore)
	elif isinstance(iterable, tuple):
		output += _callFunc("()", _tupleOut, iterable, indentationLevel, indentationString, indentBefore)
	elif isinstance(iterable, set):
		output += _callFunc("{}", _setOut, iterable, indentationLevel, indentationString, indentBefore)
	else:
		raise ValueError(
			f"Invalid value given for 'iterable'!\nReceived type '{type(iterable)}', expected one of the following:\n[dict, list, tuple, set]")
	return output


def repr_iter(iterable: Iterable, *, indentationLevel: int = 0, indentationString: str = "  ") -> str:
	r"""
	Turns iterables into a formatted string for printing.
	:param iterable:
	:param indentationLevel:
	:param indentationString:
	:return: str
	"""

	try:
		output = _iterableOut(iterable, indentationLevel, indentationString)
		return output
	except RecursionError as e:
		return ""



from os import system
from string import digits
import random
from pygetwindow import getActiveWindowTitle
from pynput.keyboard import Listener, Key, KeyCode
from .color import foreground, background

def clear() -> None:
	system("cls")

class ChoiceInterface:
	type __key = Key | KeyCode

	def __init__(
		self, *,
		textColor: foreground = foreground.WHITE,
		highlightTextColor: foreground = foreground.BLACK,
		highlightColor: background = background.WHITE,
		confirmKey: list[__key] | tuple[__key, ...] | __key = (Key.enter, Key.right),
		cancelKey: list[__key] | tuple[__key, ...] | __key = Key.esc,
		choicesSurround: str = "",
		addArrowToSelected: bool = False
	) -> None:
		r"""
		Creates the callable ChoiceInterface instance
		:param textColor:
		:param highlightTextColor:
		:param highlightColor:
		:param confirmKey:
		:param cancelKey:
		:param choicesSurround:
		:param addArrowToSelected:
		"""
		cfKey = confirmKey if isinstance(confirmKey, list) else [confirmKey]
		ccKey = cancelKey if isinstance(cancelKey, list) else [cancelKey]
		if set(cfKey) & set(ccKey):
			raise ValueError("values in confirmKey and cancelKey may not overlap!")
		self.confirmKeys = cfKey
		self.cancelKeys = ccKey
		self.textColor = textColor
		self.hlTextColor = highlightTextColor
		self.hlColor = highlightColor
		self.choicesSurround = choicesSurround
		self.addArrowToSelected = addArrowToSelected
		self.terminalWindowTitle = "Choice Interface {" + "".join(random.choices(digits, k=10)) + "}"
		self.lastKeyPressed = None

	def __call__(
			self,
			choices: list[str],
			prefix: str = "",
			suffix: str = "",
			selected: int = 0,
			minimumHighlightLength: int = -1,
			terminalTitleBefore: str = "Terminal"
		) -> int:
		r"""
		Starts the interface
		:param choices:
		:param prefix:
		:param suffix:
		:param selected:
		:param minimumHighlightLength:
		:param terminalTitleBefore:
		:return:
		"""
		system(f"TITLE {self.terminalWindowTitle}")

		if len(choices) <= 1 or (not (isinstance(choices, list) and all([isinstance(x, str) for x in choices]))):
			raise ValueError("Parameter 'lines' must be of length >= 2 and of type list[str]")
		if 0 > selected >= len(choices):
			raise ValueError(
				"Parameter 'selected' must be index of line in 'lines' and may therefore not be bigger than the "
				"biggest index of 'lines' or smaller than 0"
			)

		hlLen = minimumHighlightLength if (minimumHighlightLength >= 0 and
		                                   not (any([self.choicesSurround, self.addArrowToSelected]))) \
										else max([len(line) for line in choices]) + abs(minimumHighlightLength)

		while True:
			clear()
			if prefix:
				print(self.textColor.value + prefix+foreground.RESET.value)

			for i, line in enumerate(choices):
				_out = ''
				if i == selected:
					if not any([self.choicesSurround, self.addArrowToSelected]):
						_out = f"{self.hlColor.value+self.hlTextColor.value}{line:<{hlLen}}{foreground.RESET.value}"
					elif self.addArrowToSelected:
						_out = f"{self.hlColor.value+self.hlTextColor.value}{line+" >":<{hlLen+2}}{foreground.RESET.value}"
				else:
					_out = self.textColor.value+line+foreground.RESET.value
				if self.choicesSurround:
					_out = self.choicesSurround+_out+self.choicesSurround
				print(_out)

			if suffix:
				print(self.textColor.value+prefix+foreground.RESET.value)

			key = self._waitForKey()

			if key == Key.down and selected != len(choices)-1:
				selected += 1
			elif key == Key.up and selected != 0:
				selected -= 1
			elif key in self.confirmKeys:
				if key == Key.enter: input()
				system(f"TITLE {terminalTitleBefore}")
				return selected
			elif key in self.cancelKeys:
				if key == Key.enter: input()
				system(f"TITLE {terminalTitleBefore}")
				return -1
			elif key in [Key.down, Key.up]:
				pass
			else:
				raise Exception("Somehow, Somewhere, Something went wrong :/")

	def _waitForKey(self) -> __key | None:
		lst = Listener(on_press=lambda key: self._onKeyPress(key, lst))
		lst.start()
		lst.join()
		return self.__lastKeyPressed

	def _onKeyPress(self, key: __key | None, lst: Listener) -> None:
		if getActiveWindowTitle() != self.terminalWindowTitle:
			return
		self.__lastKeyPressed = key
		if self.__lastKeyPressed in self.confirmKeys+self.cancelKeys+[Key.up, Key.down]:
			lst.stop()


import pyperclip

def append_clipboard(text: str) -> str:
    r"""copies text

    Copies / Appends the given text to the windows clipboard using the utf-8 encoding

    Parameters
    ----------
    text: str
        The text that is copied / appended to the clipboard
    """
    pyperclip.copy(text)
    return text


import sys


def output_stdout(outputValue):
    r"""Outputs to stdout

    Outputs the given outputValue to stdout and flushes the stream after

    Parameters
    ----------
    outputValue: Any
        The value that's written to stdout
    """

    sys.stdout.write(str(outputValue) + "\n")
    sys.stdout.flush()


# -*- coding: utf-8 -*-

"""
Terminal.color
~~~~~~~~~~~~~~

Contains Variables with Color Codes to change the color in the Terminal (Windows)
"""

from ._background import background
from ._foreground import foreground



class KWArgsHandler[_VT]:
	type __allowedKeywordsType = str |\
		tuple[str, type | tuple[type, ...]] |\
		tuple[str, type | tuple[type, ...], object]

	def __init__(
			self,
			kwargs: dict[str, _VT],
			*allowedKeywords: __allowedKeywordsType
	) -> None:
		r"""
		Handles Keyword Arguments for a function
		:param kwargs:
		:param allowedKeywords:
		:raises ValueError: Invalid format for 'allowedKeywords' or disallowed keyword in 'kwargs'
		:raises TypeError: Invalid value type for keyword in 'kwargs'
		"""
		allowedKeywords_: dict[str, tuple[type | tuple[type, ...], object]] = {}
		self.kwargs = {}
		for kw in allowedKeywords:
			if isinstance(kw, tuple) and len(kw) in (2, 3):
				keyword, argType = kw[:2]
				defValue = kw[2] if len(kw) == 3 else None
				allowedKeywords_[keyword] = (argType, defValue)
			elif isinstance(kw, str):
				allowedKeywords_[kw] = (object, None)
			else:
				raise ValueError('Invalid format for allowedKeywords. \n'
				                 'Use either ("keyword", type, defaultValue) ("keyword", type) or "keyword".')
		for keyword, value in kwargs.items():
			if keyword not in allowedKeywords_:
				raise ValueError(f"Invalid keyword '{keyword}' provided.")
			if not isinstance(value, expVT := allowedKeywords_[keyword][0]):
				raise TypeError(
					f"Invalid type for keyword '{keyword}'. Expected {
						expVT.__name__ 
						if not isinstance(expVT, tuple) 
						else f"({" | ".join([vt.__name__ for vt in expVT])})"
					}, got {type(value).__name__}"
				)

		for keyword, (valueType, defValue) in allowedKeywords_.items():
			value = kwargs.get(keyword, defValue)
			if keyword not in allowedKeywords_:
				raise ValueError(f"Invalid keyword '{keyword}' provided.")
			if valueType is not None and not isinstance(value, valueType):
				raise TypeError(
					f"Invalid type for keyword '{keyword}'. Expected {
						valueType.__name__ if not isinstance(valueType, tuple) else (vt.__name__ for vt in valueType)
					}, got {type(value).__name__}.")
			self.kwargs[keyword] = value

	def __getattr__(self, attr) -> _VT:
		if attr in self.kwargs:
			return self.kwargs[attr]
		raise AttributeError(f"'KWArgsHandler' object has no attribute '{attr}'.")

	def __getitem__(self, item):
		if isinstance(item, int):
			raise IndexError("KWArgsHandler does not support indexing!")
		elif not isinstance(item, str):
			raise KeyError("KWArgsHandler only supports keys of type 'str'!")
		if item in self.kwargs:
			return self.kwargs[item]
		raise KeyError(f"KWArgsHandler object has no attribute '{item}'")

	def __repr__(self) -> str:
		return f"KWArgsHandler: {self.kwargs}"


import time


def better_print(text: str, delay: float = .01, newLine: bool = True):
    r"""Prints text

    Prints the given text letter by letter to the command line using the specified delay

    Parameters
    ----------
    text: str
        The text that is printed letter by letter

    delay: Optional[float]
        Changes the time between each printed letter
        DEFAULT: .01
    
    newLine: Optional[bool]
        whether to add a new line at the end or not
        DEFAULT: True
    """
    
    for char in text:
        print(char, end='', flush=True)
        time.sleep(delay)
    if newLine:
        print('')


from ._better_input import better_input as BetterInput
from .Terminal.color import foreground


def better_color_input(
		*args,
		sep: str = ' ',
		delay: float = .01,
		beforeColor: foreground = foreground.RESET
) -> str:
	r"""
	Prints args seperated by sep with the given delay between each character and gets an input in CYAN. Resets the
	color to `beforeColor` afterwards
	:param args:
	:param sep:
	:param delay:
	:param beforeColor:
	:return: str
	"""

	if not args:
		text = ""
	else:
		text = sep

	if text is not None:
		res = BetterInput(text, foreground.CYAN.value, delay=delay)
	else:
		res = BetterInput(foreground.CYAN.value, delay=delay)

	print(beforeColor.value, end="")

	return res


import random


class Terminal:
    r"""Functions & Stuff for the Windows Terminal

    Contains classes and Functions to use for/in the Windows Terminal

    Subclasses
    ----------
    color:
        returns escape sequences to manipulate the color of the Terminal when printed
    """

    class color:
        r"""Manipulates the color of the Terminal

        Uses special character sequences to manipulate the color of the Terminal when printed out

        Options
        -------
        RESET:
            When printed out: Restores the colors of the Terminal to the Default
        
        Subclasses
        ----------
        foreground:
            Manipulates the foreground/text color.
            Also contains Bold and Underline options for the Font

        background:
            Manipulates the background color
        """

        RESET = '\033[0m'

        class foreground:
            DEFAULT = '\033[0m'
            BOLD = '\033[1m'
            noBOLD = '\033[22m'
            UNDERLINE = '\033[4m'
            noUNDERLINE = '\033[24m'
            SWAPCOLOR = '\033[7m'
            noSWAPCOLOR = '\033[27m'
            FBLACK = '\033[30m'
            FRED = '\033[31m'
            FGREEN = '\033[32m'
            FYELLOW = '\033[33m'
            FBLUE = '\033[34m'
            FMAGENTA = '\033[35m'
            FCYAN = '\033[36m'
            FWHITE = '\033[37m'
            FDEFAULT = '\033[39m'
            FLBLACK = '\033[90m'
            FLRED = '\033[91m'
            FLGREEN = '\033[92m'
            FLYELLOW = '\033[93m'
            FLBLUE = '\033[94m'
            FLMAGENTA = '\033[95m'
            FLCYAN = '\033[96m'
            FLWHITE = '\033[97m'

            colorList = [DEFAULT, FRED, FGREEN, FYELLOW, FBLUE, FMAGENTA, FCYAN, FWHITE, FDEFAULT, 
                         FLBLACK, FLRED, FLGREEN, FLYELLOW, FLBLUE, FLMAGENTA, FLCYAN, FLWHITE]
            def random():
                colorList = Terminal.color.foreground.colorList
                return random.choice(colorList)

        class background:
            DEFAULT = '\033[0m'
            SWAPCOLOR = '\033[7m'
            noSWAPCOLOR = '\033[27m'
            BBLACK = '\033[40m'
            BRED = '\033[41m'
            BGREEN = '\033[42m'
            BYELLOW = '\033[43m'
            BBLUE = '\033[44m'
            BMAGENTA = '\033[45m'
            BCYAN = '\033[46m'
            BWHITE = '\033[47m'
            BDEFAULT = '\033[49m'
            BLBLACK = '\033[100m'
            BLRED = '\033[101m'
            BLGREEN = '\033[102m'
            BLYELLOW = '\033[103m'
            BLBLUE = '\033[104m'
            BLMAGENTA = '\033[105m'
            BLCYAN = '\033[106m'
            BLWHITE = '\033[107m'

            colorList = [DEFAULT, BBLACK, BRED, BGREEN, BYELLOW, BBLUE, BMAGENTA, BCYAN, BWHITE,
                         BDEFAULT, BLBLACK, BLRED, BLGREEN, BLYELLOW, BLBLUE, BLMAGENTA, BLCYAN, BLWHITE]
            def random():
                colorList = Terminal.color.background.colorList
                return random.choice(colorList)


class kdict(dict):
	r"""Like a dict but accessible with x.key and x["key"]

	A class that creates a dict that is also accessible by using x.key, not just by using x["key"]. Only works with keys that pass str.isidentifier()!

	How To Use
	----------
	x = kdict({"key": value, ...}); y = x.key
	print(y) # Output: value

	OR

	d = {"key": value, ...}; x = kdict(d); y = x.key
	print(y) # Output: value
	"""
	def __getattr__(self, name:str):
		if name in self and name.isidentifier():
			return self[name]
		raise KeyError(f"'{name}'")
	
	def __setattr__(self, name, value) -> None:
		self[name] = value


from functools import wraps
from inspect import signature, Parameter
from typing import Callable


def check_params[P, R](func: Callable[[P], R]) -> Callable[[P], R]:
	r"""
	Checks that all given parameters are the expected type. Excludes **kwargs!
	:param func:
	:raises TypeError: Parameter's value is not of the expected type.
	:return: The decorated function.
	"""
	@wraps(func)
	def wrapper(*args, **kwargs) -> R:
		params = signature(func).parameters
		for arg, param in zip(args, params.values()):
			annot = param.annotation
			if not isinstance(arg, annot) and annot != Parameter.empty:
				raise TypeError(
					f"Invalid type for parameter {param.name}: {type(arg)}. Expected: {annot}"
				)

		for k, v in kwargs.items():
			if k not in params:
				continue
			annot = params[k].annotation
			if not isinstance(v, annot) and annot != Parameter.empty:
				raise TypeError(
					f"Invalid type for parameter '{k}': {type(v)}. Expected: {params[k].annotation}"
				)

		return func(*args, **kwargs)
	return wrapper


from .better_print import better_print as BetterPrint


def better_input(text: str = "", delay: float = .01):
    r"""Gets an input

    Prints the given text letter by letter using the specified delay and gets an input() after

    Parameters
    ----------
    text: Optional[str]
        The text that is printed letter by letter
        DEFAULT: ""

    delay: Optional[float]
        Changes the time between each printed letter
        DEFAULT: .01
    """
    
    if text:
        BetterPrint(text, delay, False)
    res = input()
    return res


# -*- coding: utf-8 -*-

"""
Maxi's Functions
~~~~~~~~~~~~~~~~

Contains all my Functions

"""

from .append_clipboard          import append_clipboard			as appendClipboard
from .better_color_input        import better_color_input       as betterColorInput
from .better_input              import better_input             as betterInput
from .better_print              import better_print             as betterPrint
from .color_input               import color_input              as colorInput
from .key_dict					import kdict					as kdict
from .kwargs_handler			import KWArgsHandler			as KWArgsHandler
from .output_stdout             import output_stdout            as outputStdout
from .terminal                  import Terminal                 as Terminal


from .Terminal.color import foreground

def color_input(text: str = "", beforeColor: foreground = foreground.RESET):
    r"""
    Gets an input in CYAN. Resets the color to `beforeColor` afterwards
    :param text:
    :param beforeColor:
    :return:
    """
    if text:
        res = input(text + foreground.CYAN.value)
    else:
        res = input(foreground.CYAN.value)
    if beforeColor is not None:
        print(beforeColor.value, end="")
    return res


class kdict(dict):
	r"""Like a dict but accessible with x.key and x["key"] if key passes str.isidentifier()"""

	def __getattr__(self, name: str):
		if name in self and name.isidentifier():
			return self[name]
		raise KeyError(f"'{name}'")
	
	def __setattr__(self, name, value) -> None:
		self[name] = value


from typing import Optional
from .terminal import Terminal

def color_input(text: str = "", beforeColor: Optional[Terminal.color.foreground] = Terminal.color.RESET):
    r"""Gets an input

    executes the input() function in cyan and changes the color back to beforeColor if given

    Parameters
    ----------
    text: Optional[str]
        The text that is printed before getting the input
        DEFAULT: ""
    beforeColor: Optional[Terminal.color.foreground]
        The color that was used before (if u want it to change back)
        DEFAULT: Terminal.color.RESET
    """
    if text:
        res = input(text + Terminal.color.foreground.FCYAN)
    else:
        res = input(Terminal.color.foreground.FCYAN)
    if beforeColor is not None:
        print(beforeColor, end="")
    return res


from typing import Optional
from .better_input import better_input as BetterInput
from .terminal import Terminal


def better_color_input(text: str = None, beforeColor: Optional[Terminal.color.foreground] = Terminal.color.RESET, delay: float = .01):
    r"""Gets an input

    Prints the given text letter by letter using the specified delay and gets an input() after

    Parameters
    ----------
    text: Optional[str]
        The text that is printed letter by letter
        DEFAULT: None
    
    beforeColor: Optional[Terminal.color.foreground]
        The color to change back to after getting the input
        DEFAULT: Terminal.color.RESET

    delay: Optional[float]
        Changes the time between each printed letter
        DEFAULT: .01
    """
    if text is not None:
        res = BetterInput(text, Terminal.color.foreground.FCYAN, delay)
    else:
        res = BetterInput(end=Terminal.color.foreground.FCYAN, delay=delay)
    if beforeColor is not None:
        print(beforeColor, end="")
    return res


from typing import Type, Tuple, Union

class KWArgsHandler:
	r"""Handles allowed keywords and types for values in kwargs

	Makes sure the parsed `kwargs` only contain allowed keywords and the correct types of values for the corresponding keywords

	Parameters
	----------
	kwargs: dict
		The keyword arguments to be used
	
	*allowedKeywords: Union[str, tuple(str, type, any)]
		tuple:
			0: the allowed keyword
			1: the expected type of value (object for any)
			2: the default value, if not given: None
		str: the allowed keyword
	"""
	def __init__(self, kwargs: dict, *allowedKeywords: Tuple[Union[str, Tuple[str, Type[object], object]]]) -> None:
		allowedKeywords_ = {}
		self.kwargs = {}
		for kw in allowedKeywords:
			if isinstance(kw, tuple) and len(kw) in (2, 3):
				keyword, argType = kw[:2]
				defValue = kw[2] if len(kw) == 3 else None
				allowedKeywords_[keyword] = (argType, defValue)
			elif isinstance(kw, str):
				allowedKeywords_[kw] = (object, None)
			else:
				raise ValueError("Invalid format for allowedKeywords. Use either (\"keyword\", type) or \"keyword\".")
		for keyword, (valueType, defValue) in allowedKeywords_.items():
			value = kwargs.get(keyword, defValue)
			if keyword not in allowedKeywords_:
				raise ValueError(f"Invalid keyword '{keyword}' provided.")
			if valueType is not None and not isinstance(value, valueType):
				raise TypeError(f"Invalid type for keyword '{keyword}'. Expected {valueType.__name__}, got {type(value).__name__}.")
			self.kwargs[keyword] = value

	def __getattr__(self, name):
		if name in self.kwargs:
			return self.kwargs[name]
		raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'.")
	
	def __repr__(self) -> str:
		return f"KWArgsHandler: {self.kwargs}"